//#include <iostream.h>
//#include "dyna.cxx"

class BitField {
  long bitLength, length;
  FastDynaArray<unsigned long> data;
public:
  BitField() {
    #ifdef LOGGING
    cout << "BitField (" << this << "): default constructor\n";
    #endif
  }
  BitField(long bits) {
    #ifdef LOGGING
    cout << "BitField (" << this << "): parameterised constructor (" << bits << " bits)\n";
    #endif
    setLength(bits);
  }
  ~BitField() {}  

  void check(long bit) {
    int old = data.getLength();
    int bits = bit + 1;
    int target = (bits >> 5) + ( (bits % 32) ? 1 : 0 );
    data.check(target);
    length = data.getLength();
    if (length > old) {
      for (int n=old; n<length; n++)
	data[n] = 0;
    }
    bitLength = length*32;
  }

  void setLength(long bits) {
    bitLength = bits;
    length = (bits >> 5) + ( (bits % 32) ? 1 : 0 );
    #ifdef LOGGING
    cout << "BitField (" << this << "): setting length to " << bits << " bits (" << (length<<2) << " bytes)\n";
    #endif
    data.setLength(length);
  }

  void clear() {
    for (long l=0; l<length; l++) data[l] = 0;
  }

  int output(char *filename) {
    FILE *fp = fopen(filename, "wb");
    if (!fp) {
      cout << "Bitfield (" << this << "): cannot open file '" << filename << "' for output." << endl;
      return -1;
    }
    return output(fp);
  }

  int input(char *filename) {
    FILE *fp = fopen(filename, "rb");
    if (!fp) {
      cout << "Bitfield (" << this << "): cannot open file '" << filename << "'. for input" << endl;
      return -1;
    }
    return input(fp);
  }

  int input(FILE *fp)  {
    fread(&bitLength, sizeof(long), 1, fp);
    setLength(bitLength);
    return fread(data.getArray(), sizeof(unsigned long), length, fp);
  }

  int output(FILE *fp) {
    #ifdef LOGGING
    cout << "Bitfield (" << this << "): output size " << length << endl;
    #endif
    fwrite(&bitLength, sizeof(long), 1, fp);
    return fwrite(data.getArray(), sizeof(unsigned long), length, fp);
  }

  inline int operator()(long index) const {
    return (data(index>>5) & (1 << (index % 32))) ? 1 : 0;
  }

  inline void set1(long index) {
//    cout << "set1 in (" << index << ")\n";
    data[index >> 5] |= (1 << (index % 32));
//    assert((data(index>>5) & (1 << (index % 32)))!=0); // ###
//    cout << "set1 out\n";
  }
  inline void set0(long index) {
//    cout << "set0 in (" << index << ")\n";
//    cout << "DATA BEFORE: " << setbase(16) << data[index >> 5] << endl;
    data[index >> 5] &= (~(1 << (index % 32)));
//    cout << "MASK: " << ~(1 << (index % 32)) << endl;
//    cout << "DATA AFTER: " << data[index >> 5] << setbase(10) << endl;
//    assert((data(index>>5) & (1 << (index % 32)))==0); // ###
//    cout << "set0 out\n";
  }
};

