////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

// Update History: 26 April 1996 (setLength() memory leak fixed)
//                  6 May   1996 (bounds checking )

#ifndef __DYNA_CXX
#define __DYNA_CXX

//#define VERBOSE
#include "dyna.hpp"

template <class T>
FastDynaArray<T>::FastDynaArray()
{
  length = 0;
  array = NULL;
}

template <class T>
FastDynaArray<T>::FastDynaArray(long size)
{
#ifdef VERBOSE
  cout << "FastDynaArray (" << this << "): parameterised constructor (size " << size << ")\n";
  if (size < 0) cout << "FastDynaArray: Warning! Initial size < 0\n";
#endif

  if (size <= 0) {
    length = 0;
    array = NULL;
    return;
  }

  array = new T[size];
  if (!array) {
    cout << "FastDynaArray: (" << this << "): memory allocation (size " << size << ") failed!\n";
    length = 0;
    array = NULL;
    return;
  }

  length = size;
  return;
}

template <class T>
int FastDynaArray<T>::expand(long index)
{
#ifdef VERBOSE
  cout << "FastDynaArray (" << this << "): expanding to accomodate index " << index << "\n";
#endif
  long targetLength = 1;
  while (index >= targetLength) targetLength <<= 1;
  if (!adjust(targetLength - length)) return 0;
  return 1;
}

/*
template <class T>
int FastDynaArray<T>::expand(long index, const T value)
{
#ifndef DYNASILENT
  cout << "FastDynaArray (" << this << "): expanding to accomodate index " << index << "\n(assigning " << value << " to new elements)";
#endif
  long targetLength = 1, oldLength = length;
  while (index >= targetLength) targetLength <<= 1;
    if (!adjust(targetLength - index)) return 0;
  for (long n=length; n<targetLength; n++) array[n] = value;
  return 1;
}
*/

/* template <class T>
long FastDynaArray<T>::search (T target) const
{
#ifndef DYNASILENT
  cout << "FastDynaArray (" << this << "): searching for [" << target << "]\n";
#endif
  for (long n=0; n<length; n++)
    if (array[n] == target) return n;
  return -1;
} */

template <class T>
int FastDynaArray<T>::adjust(long size)
{
#ifdef VERBOSE
  cout << "FastDynaArray (" << this << "): adjusting by " << size << " elements to " << size+length << " elements\n";
#endif
  
  long newLength = length + size;
  if (newLength <= 0) {
    if (length && array) delete [] array;
    array = NULL;
    length = 0;
    #ifdef VERBOSE
      cout << "FastDynaArray (" << this << "): array size set to 0 elements\n";
    #endif
    return 1;
  }

  T *extendedArray = new T[size+length];
  if (!extendedArray) {
    cout << "FastDynaArray (" << this << "): array extension (size " << size << ") failed!\n";
    return 0;
  }

  if (size >= 0)
    for(long i=0; i<length; i++) extendedArray[i] = array[i];
  else {
    for(long i=0; i<newLength; i++) extendedArray[i] = array[i];
  }

  if (length && array) delete [] array;

  length = newLength;
  array = extendedArray;

#ifdef VERBOSE
  cout << "FastDynaArray (" << this << "): array expanded to " << length << " elements\n";
#endif
  return 1;
}

template <class T>
FastDynaArray<T>::~FastDynaArray()
{
#ifdef VERBOSE
  long elements = length;
  cout << "FastDynaArray (" << this << "): destructor\n";
#endif
  if (array) delete [] array;
  length=0;
#ifdef VERBOSE
  cout << "FastDynaArray (" << this << "): destroyed (" << elements << " elements)\n";
#endif
}

template <class T>
void FastDynaArray<T>::destroy()
{
  //  cout << "FastDynaArray (" << this << "): destroying " << length << " elements " << endl;
  delete [] array;
  array = NULL;
  length = 0;
}


template <class T>
ostream &operator<<(ostream &os, const FastDynaArray<T> &dyna)
{
  for(long n=0;n<dyna.length;n++)
    os << n << ": " << dyna(n) << " ";
  os << endl;
  return os;
}
#endif

