#ifndef __DYNA_HPP
#define __DYNA_HPP

/////////////////////////////////////////////////////////////////
//             Generic Fast Dynamic Array Class                //
/////////////////////////////////////////////////////////////////

template <class T>
class FastDynaArray
{
  protected:
    long length;
    T *array;
    int expand(long index);
//    expand(long index, const T value);
  public:
    FastDynaArray();
    FastDynaArray(long size);
    ~FastDynaArray();

    inline long getLength() const { return length; }
    inline int setLength(long size) {
#ifdef VERBOSE
      cout << "FastDynaArray (" << this << "): setting length to " << size << " elements\n";
#endif
      if (size == length) return 1;
      if (!adjust(size - length)) return 0;
      length = size;
      return 1;
    }

    long search(T target) const;
    inline T *getArray() const { return array; }
    inline void setArray(T *newArray, long size) { array = newArray; length = size; }

    inline T &operator[] (register long index) {
#ifdef SAFE
      if ( (index >= length) || (index < 0) ) {
	cout << "FastDynaArray: (" << this << ") Array subscript [" << index
	     << "] out of range (Valid range is 0 to " << length - 1 <<
          ")" << endl;
        assert(index >=0 && index < length);
      }
#endif
      return array[index];
    }

    inline T &operator() (register long index) const {
#ifdef SAFE
      if ( (index >= length) || (index < 0) ) {
	cout << "FastDynaArray: Array subscript [" << index
	     << "] out of range (Valid range is 0 to " << length - 1 << 
          ")" << endl;
	assert(index >=0 && index < length);
      }
#endif
      return array[index];
    }

    inline void check(register long index) {
      if (index >= length) expand(index);
    }

    inline void assign(long index, const T value) {
      if (index >= length) expand(index);
      array[index] = value;
    }
    int adjust(long size);
  void destroy();

    //friend ostream &operator << <T>(ostream &os, const FastDynaArray<T> &dyna);
};
#endif

