///////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

/*
 This file implements some filename handling functions
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


/*
  Remove the extension from the filename
*/
char* remove_extension ( char* orig_name )
{
  int length, new_length, last_dot = 0, i ;
  char* new_string ;

  length = strlen (orig_name) ;
  for ( i = 0; i < length ; i++ )
    {
      if (orig_name[i] == '.')
	last_dot = i ;
    }

  if (last_dot == 0)
    new_length = length ;
  else
    new_length = last_dot ;

  new_string = (char*) malloc (new_length) ;

  strncpy (new_string, orig_name, new_length) ;
  new_string[new_length] = '\0' ;

  return new_string ;
}


/*
   Adds an extension to a file name producing a new file
   name which has not modified the original
*/
char* add_extension (char* orig_name, char* new_extension)
{
   int length = 0, ext_length = 0 ;
   char* new_string ;

   ext_length = strlen (new_extension) ; 
   length = strlen (orig_name) + ext_length + 2 ;
   
   new_string = (char*) malloc (length) ;
   new_string = strcpy (new_string, orig_name) ;
   new_string = strncat(new_string, ".", 1) ;
   new_string = strncat(new_string, new_extension, ext_length) ;
   new_string[length - 1] = '\0' ;

   return new_string ;
}

/*
   Adds a preceding "./" to the name if the name does not start with
   a "/" (which indicates it's not an absolute path)
*/
char* add_relative_start (char* orig_name )
{
   if (orig_name[0] == '/')
      return orig_name ;
   int length = strlen(orig_name) ;
   /* 3 -- two for "./" and one for the null terminating \0 */
   char* new_name = (char*) malloc (length + 3);
   
   new_name[0] = '.' ;
   new_name[1] = '/' ;
   for (int i = 0; i < length; i++)
      new_name[i+2] = orig_name[i] ;
   new_name[length+2] = '\0' ;

   return new_name ;
}

