////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
// #include <pwd.h>
#include <cstring>
#include <cassert>
#include <cctype>
#include <ctime>
#include "common.hpp"
#include "dyna.hpp"
#include "info.hpp"
#include "token.hpp"
#include "parse.hpp"

void Parser::headers(ostream &user) 
{
  // char buffer[512], *creator, *login;
  // struct passwd *userDetails;
  time_t timeNow = time(NULL);

  //  userDetails = getpwuid(getuid());
  //  creator = userDetails->pw_gecos;
  // login = userDetails->pw_name;

  static char *includes[] = { "<iostream>", "<memory>", "<cstdio>", "<cmath>", "<complex>", "<cstdlib>", "\"dyna.hpp\"", "\"common.hpp\"", "\"element.hpp\"", "\"user.hpp\"", NULL };

  static char *methods[] = {
    "State() {\n  __kids = 0;\n#ifdef STATE_COUNT\n  __instances++;\n#endif\n#ifdef SAFE\n  __primaryKey = __secondaryKey = -1;\n#endif\n  } //",
    "~State() {\n#ifdef STATE_COUNT\n  __destroyed++;\n#endif\n  } //",
    "void initial()",
    "int isEnabled(int *__en) const",
    "int isTarget() const",
    "int isExcluded() const",
    "double passageWeight() const",
    "void fire(int __t, State *next) const",
    "void calcHelpValues()",
    "void computeStateMeasures(double *) const",
    "void computeCountMeasures(int *, int, double *) const",
    "double getTransWeight(int __t) const",
    "void checkInvariants() const",
    "int input(FILE *)",
    "int output(FILE *) const",
    "void output() const",
    "unsigned long calcPrimaryHashKey() const",
    "unsigned long calcSecondaryHashKey() const",
    "long size() const",
    "void calcHashKeys(unsigned long &__primary, unsigned long &__secondary)",
    "inline void computeHashKeys() {    calcHashKeys(__primaryKey, __secondaryKey); } //",
    "inline StateType getType() const { return __type; } //",
    "inline unsigned long getPrimaryKey() const { return __primaryKey; } //",
    "inline unsigned long getSecondaryKey() const { return __secondaryKey; } //",
    "int findEnabledTransitions()",
    "inline int getEnabledTransition(int __t) { return __enabledTrans(__t); } //",
    "inline void fireEnabledTransition(int __t, State *next) const { fire(__enabledTrans(__t), next); } //",
    "inline int numberChildren() const { return __kids; } //",
    "inline int getNumberEnabled() const { return __numberEnabled; } //",
    "inline void setKids(int __k) { __kids = __k; } //",
    "inline double getEnabledTransWeight(int __t) const { return getTransWeight(__enabledTrans(__t)); } //",
    "inline StateMatrixElement &getChild(int __t) const { return __children(__t); } //",
    "inline double rateOut() { double __sum = 0; for (int __n=0; __n<__kids; __n++) __sum += __children(__n).getWeight(); return __sum; } //",
    "inline double getTotalTransWeight() const {\n    double __sum = 0;\n    for (int __t=0; __t<getNumberEnabled(); __t++)\n      __sum += getEnabledTransWeight(__t);\n    return __sum;\n  } //",
    "inline void setChild(int __t, long __value, double __weight) {\n    if (__weight <= 0) {\n      output();\n      cout << endl;\n      return;\n    }\n    for (long __n=0; __n<__kids; __n++) {\n      if (__children(__n).getDestination()==__value) {\n        __children[__n].setWeight(__children(__n).getWeight()+__weight);\n        return;\n      }\n    }\n    __children.check(__kids);\n    __children[__kids].setDestination(__value);\n    __children[__kids].setWeight(__weight);\n    if (__kids == 255 ) { printf (\"Sorry too many kids\\n \"); exit(1) ; } ; __kids++;\n  } //",
    "inline int matrixOutput(FILE *__mf) {\n    int __k = __kids;\n    /*printf(\"hello again __k = \%d\\n\", __k); */\n     fwrite(&__k, sizeof(int), 1, __mf);\n    return fwrite(__children.getArray(), sizeof(StateMatrixElement), __kids, __mf);\n  } //",
    "inline void calculateCountMeasures(double *__measures) const {\n    computeCountMeasures(__enabledTrans.getArray(),\n      __numberEnabled, __measures);\n  } //",
    "int initialise(int, TransType *,int *)",
    NULL
  };

  long n;

  user << "// HYDRA version " << VERSION << " (" << DATE << ")" << endl;
  user << "// " << COPYRIGHT << endl;
  user << endl;
  user << "// Automatically generated from '" << sourceFile << "' on " << ctime(&timeNow) << endl;

  //  user << "#define real double" << endl;
  for (n = 0; n<constants; n++)
    user << "#define " << cons(n)->getIdentity() << " " << cons(n)->getDefine() << endl << endl;

  user << "using namespace std;" << endl << endl;

  for (n = 0; includes[n]; n++)
    user << "#include " << includes[n] << endl;
  if (includeCB.getArray()) user << includeCB << endl;
  user << endl;

  user << "class State {" << endl;
  user << "private:" << endl <<"  // state vector description" << endl;

  char last[MAX_LENGTH];
  strcpy(last, "");
  for (n = 0; n < elements; n++) {
    char *current = instance(n);
    char *_type = type(n);
    if (!strcmp(last, _type)) {
      if (last[strlen(last) - 1] == '*')
        user << ", *" << current;
      else
	user << ", " << current;
    } else {
      if (n) user << ";" << endl;
      if (strlen(_type) && _type[strlen(_type) - 1] == '*')
        user << "  " << _type << current;
      else 
	user << "  " << _type << " " << current;
    }
    strcpy(last, _type);
  }
  if (elements) user << ";" << endl;

  strcpy(last, "");

  user << "  // ### do not change the relative offset of kids" << endl;
  user << "  unsigned char __kids;" << endl << endl;
  user << "  static int __trans;" << endl;
  user << "  static int *__enabled, *__priority;" << endl;
  user << "  static TransType *__transitionType;" << endl << endl;
  user << "  unsigned long __primaryKey, __secondaryKey;" << endl;
  user << "  StateType __type;" << endl << endl;
  user << "  int __numberEnabled;" << endl;
  user << "  FastDynaArray<int> __enabledTrans;" << endl << endl;
  user << "  FastDynaArray<StateMatrixElement> __children;" << endl << endl;

  if (helpValues) {
    user << endl << "  // help values" << endl;
  }
  for (n = 0; n < helpValues; n++) {
    char *current = helpv(n)->getIdentity();
    char *type = helpv(n)->getType();
    if (!strcmp(last, helpv(n)->getType() )) {
      if (last[strlen(last) - 1] == '*')
        user << ", *" << current;
      else
	user << ", " << current;
    } else {
      if (n) user << ";" << endl;
      if (strlen(type) && type[strlen(type) - 1] == '*')
	user << "  " << type << current;
      else
	user << "  " << type << " " << current;
    }
    strcpy(last, type);
  }
  if (helpValues) user << ";" << endl;

//  user << stateVectorCB << endl;

  user << "public:" << endl;
  for (n = 0; methods[n]; n++)
    user << "  " << methods[n] << ";" << endl;
  user << "};" << endl << endl;
}

void Parser::declarations(ostream &user) {
  user << endl << "TransType *State::__transitionType = NULL;" << endl;
  user << "int State::__trans = 0;" << endl;
  user << "int *State::__enabled = NULL;" << endl;
  user << "int *State::__priority = NULL;" << endl << endl;
}

void Parser::calcHasKeys(ostream &user) {
  user << "void State::calcHashKeys(unsigned long &__primary, unsigned long &__secondary) {" << endl;
  user << "  __primary = calcPrimaryHashKey();" << endl;
  user << "  __secondary = calcSecondaryHashKey();" << endl;
  user << "}" << endl << endl;
}
 
void Parser::initialise(ostream &user) {
  user << "int State::initialise(int __transitions, TransType *__transType, int *__transPriority) {" << endl;
  user << "  __trans = __transitions;" << endl;
  user << "  __transitionType = new TransType[__trans];" << endl;
  user << "  __enabled = new int[__trans];" << endl;
  user << "  __priority = new int[__trans];" << endl;
  user << "  for (int __n=0; __n<__trans; __n++) {" << endl;
  user << "    __transitionType[__n] = __transType[__n];" << endl;
  user << "    __priority[__n] = __transPriority[__n];" << endl;
  user << "  }" << endl;
  user << "  return 1;" << endl;
  user << "}" << endl << endl;
}

void Parser::findEnabledTransitions(ostream &user) {
  user << "int State::findEnabledTransitions() {" << endl;
  user << "  int __vanishing = 0;" << endl;
  user << "  long ___timedPriority = 0, __instantPriority = 0;" << endl;
  user << "  __numberEnabled = 0;" << endl;
  user << "  __type = TANGIBLE;" << endl;
  user << "  isEnabled(__enabled);" << endl;
  user << "  for (int __t=0; __t<__trans; __t++) {" << endl;
  user << "    switch(__transitionType[__t]) {" << endl;
  user << "    case TIMED:" << endl;
  user << "      if (!__vanishing && __enabled[__t] && __priority[__t] >= ___timedPriority) {" << endl;
  user << "        if (__priority[__t] > ___timedPriority) {" << endl;
  user << "          __numberEnabled = 0;" << endl;
  user << "          ___timedPriority = __priority[__t];" << endl;
  user << "	   }" << endl;
  user << "        __enabledTrans.check(__numberEnabled);" << endl;
  user << "        __enabledTrans[__numberEnabled] = __t;" << endl;
  user << "        __numberEnabled++;" << endl;
  user << "      }" << endl;
  user << "      break;" << endl;
  user << "    case INSTANTANEOUS:" << endl;
  user << "      if (__enabled[__t] && __priority[__t] >= __instantPriority) {" << endl;
  user << "        if (!__vanishing) {" << endl;
  user << "          __type = VANISHING;" << endl;
  user << "          __numberEnabled = 0;" << endl;
  user << "          __vanishing = 1;" << endl;
  user << "         __instantPriority = __priority[__t];" << endl;
  user << "        } else if (__priority[__t] > __instantPriority) {" << endl;
  user << "	     __numberEnabled = 0;" << endl;
  user << "	     __instantPriority = __priority[__t];" << endl;
  user << "        }" << endl;
  user << "        __enabledTrans.check(__numberEnabled);" << endl;
  user << "        __enabledTrans[__numberEnabled] = __t;" << endl;
  user << "        __numberEnabled++;" << endl;
  user << "        break;" << endl;
  user << "      }" << endl;
  user << "    }" << endl; 
  user << "  }" << endl;
  user << "  __enabledTrans.setLength(__numberEnabled);" << endl;
  user << "  return __numberEnabled;" << endl;
  user << "}" << endl << endl;
}

void Parser::initial(ostream &user) {
  user << "void State::initial() { // initial state vector" << endl;
  user << initialCB << endl;
  if (helpValues) user << "  calcHelpValues();" << endl;
  if (invariants) user << "  checkInvariants();" << endl;
  user << "}" << endl << endl;
}

void Parser::isEnabled(ostream &user) {
  user << "int State::isEnabled(int *__enabled) const {" << endl;
  for (int n=0; n<transitions; n++)
    user << "  __enabled[" << setw(4) << n << "] = (" << trans[n]->getCondition() << "); " << endl;
  user << "  return 1;" << endl;
  user << "}" << endl << endl;
}

void Parser::isTarget(ostream &user) {
  user << "int State::isTarget() const {" << endl;

  if (targetCB.getArray())
    user << "  return (" << targetCB << ");" << endl;
  else
    user << "  return 0;" << endl;
  user << "}" << endl << endl;
}


void Parser::isExcluded(ostream &user) {
  user << "int State::isExcluded() const {" << endl;

  if (excludedCB.getArray())
    user << "  return (" << excludedCB << ");" << endl;
  else
    user << "  return 0;" << endl;
  user << "}" << endl << endl;
}

void Parser::passageWeight(ostream &user) {
  user << "double State::passageWeight() const {" << endl;

  /*  
  if (weightCB.getArray())
    user << weightCB << endl;
  else
    user << "  return 1;" << endl;
  */

  if (weightCB.getArray())
    user << "  return (" << weightCB << ");" << endl;
  else
    user << "  return 0;" << endl;


  user << "}" << endl << endl;
}

void Parser::calcHelpValues(ostream &user) {
  user << "void State::calcHelpValues() {" << endl;
  for (int n=0; n<helpValues; n++) {
    user << "  " << helpv(n)->getIdentity() << " = " << helpv(n)->getExpression() << ";" << endl;
  }
  user << "}" << endl << endl;
}

void Parser::checkInvariants(ostream &user) {
  user << "void State::checkInvariants() const {" << endl;
  for (int n=0; n<invariants; n++) {
    user << "  if (!(" << invar(n)->getExpression() << ")) {" << endl;
    user << "    cout << \"warning: invariant (" << invar(n)->getExpression() << ") failed on state:\" << endl;" << endl;
    user << "    output();" << endl;
    user << "  }" << endl;
  }
  user << "}" << endl << endl;
}

void Parser::fire(ostream &user) {
  user << "void State::fire(int __t, State *next) const {" << endl;
  user << "  memcpy(next, this, size());" << endl;
  user << "  switch(__t) { // transition actions" << endl;
  for (int n=0; n<transitions; n++) {
    user << "  case " << setw(4) << n << ":" << endl;
    user << trans[n]->getAction() << endl;
    user << "    break;" << endl;
  }
  user << "  default  :" << endl;
  user << "    cout << \"fire() error: transition \" << __t << \" out of range!\" << endl;" << endl;
  user << "  }" << endl;
  if (helpValues) user << "  next->calcHelpValues();" << endl;
  if (invariants) user << "  next->checkInvariants();" << endl;
  user << "}" << endl << endl;
}

void Parser::getTransWeight(ostream &user)
{
  user << "double State::getTransWeight(int __t) const" << endl;
  user << "{" << endl;
  user << "  switch (__t) {" << endl;
  for (int n=0; n<transitions; n++)
    user << "  case " << setw(4) << n << ": return " << trans[n]->getWeight() << ";" << endl;
  user << "  default  :" << endl;
  user << "    cout << \"getTransWeight() error: transition \" << __t << \" out of range!\" << endl;" << endl;
  user << "  }" << endl;
  user << "  return 0.0;" << endl;
  user << "}" << endl << endl;  
}

void Parser::computeStateMeasures(ostream &user)
{
  user << "void State::computeStateMeasures(double *__measures) const" << endl;
  user << "{" << endl;
  for (int n=0; n<stateMeasures; n++) {
    user << "  __measures[" << setw(4) << n << "] = " << statem[n]->getExpression() << ";" << endl;
  }
  user << "}" << endl << endl;
}

int Parser::nonTrivialCondition(CodeBlock &condition)
{
  if (!condition.getArray()) return 0;
  if (!strlen(condition.getArray()) || !strcmp("1", condition.getArray()))
    return 0;
  return 1;
}

void Parser::computeCountMeasures(ostream &user)
{
  user << "void State::computeCountMeasures(int *__enabledTrans, int __numberEnabled, double *__measures) const" << endl;
  user << "{" << endl;
  user << "  int __trans, __t;" << endl;
  user << "  static State *__next = new State;" << endl;
  user << "  for (__trans = 0; __trans < __numberEnabled; __trans++) {" << endl;
  user << "    __t = __enabledTrans[__trans];" << endl;

  int n, t, code, trans, list = 0;

  int fire = 0;
  for (n=0; n < countMeasures; n++)
    if (nonTrivialCondition(countm[n]->getPostcondition())) fire = 1;
  if (fire) user << "    fire (__t, __next);" << endl;

  CountMeasure *c;
  for (n=0; n<countMeasures; n++) {
    c = countm[n];
    trans = c->getTrans();
    code = nonTrivialCondition(c->getPrecondition());
    if (nonTrivialCondition(c->getPostcondition())) code |= 2;
    if (trans && !c->getAll()) list = 1;
    user << "    ";
    switch (code) {
    case 0: // neither pre, nor post
      if (list) user << "if (";
      break;
    case 1: // pre condition only
      user << "if (" << c->getPrecondition();
      break;
    case 2: // postcondition only
      user << "if (" << c->getPostcondition();
      break;
    case 3: // pre and post conditions
      user << "if (" << c->getPrecondition() << " && " << c->getPostcondition();
      break;
    default:
      cout << "(impossible count measure code)" << endl;
      break;
    }
    if (!list) {
      if (code) user << ") " << endl << "      ";
    } else {
      if (code) user << " && (";
      for (t=0; t<trans; t++) {
        if (t) user << " || ";
        user << "__t == " << c->getTransition(t);
      }
      if (code) user << ")) "; else user << ") ";
      user << endl << "      ";
    }
    user << "__measures[" << setw(4) << n << "] += getTransWeight(__t);" << endl;
  }
  user << "  }" << endl;  
  user << "  // delete next;" << endl;
  user << "}" << endl << endl;
}

void Parser::output(ostream &user)
{
  user << "void State::output() const" << endl;
  user << "{" << endl;
  if (outputCB.getArray())
    user << outputCB << endl;
  else {
    user << "  cout << \"(\" << ";
    for (int n=0; n < elements; n++) {
      user << instance(n) << " << ";
      if (n < elements - 1)
        user << "\",\" << ";
    }
    user << "\")\" << endl;" << endl;
  }
  user << "}" << endl << endl;
}

void Parser::primaryHashKey(ostream &user)
{
  long slide, top;
  if (elements > 6) {
    slide = 2;
    top = 14;
  } else {
    slide = 16/elements;
    top = 16-slide;
  }
  user << "unsigned long State::calcPrimaryHashKey() const" << endl;
  user << "{" << endl;
  if (primaryCB.getArray()) {
    user << primaryCB << endl;
  } else {
    if (advancedType) {
      tokeniser->error("use of advanced types in state vector requires overloading of primary hash function");
      user << "  return 0; " << endl << "}" << endl << endl;
      return;
    }
    int factor = 0;
    int shift = 0;
    user << "  register unsigned long __key = ";
    if (!elements) user << "1;" << endl;
    for (int n=0; n<elements; n++) {
      if (shift + factor) 
        user << "(" << instance(n) << " << " << shift + factor << ")";
      else
        user << instance(n);
      if (n < elements - 1) user << " ^ ";
        else user << ";" << endl;
      shift += slide;
      if (shift > top) {
        shift = 0;
        factor = !factor;
      }
    }
    user << "  return (__key % 350003);" << endl;
  }
  user << "}" << endl << endl;
}

void Parser::secondaryHashKey(ostream &user)
{
  user << "unsigned long State::calcSecondaryHashKey() const" << endl;
  user << "{" << endl;
  if (secondaryCB.getArray()) {
    user << secondaryCB << endl;
  } else {
    if (advancedType) {
      tokeniser->error("use of advanced types in state vector requires overloading of secondary hash function");
      user << "  return 0; " << endl << "}" << endl << endl;
      return;
    }
    if (!elements) {
      user << "return 1;" << endl;
    } else {
      if (safeHash)
	user << "  static long ___p[] = { 32768, 16384, 8192, 4096, 2048, 1024, 512, 256, 128, 64, 32, 16, 8, 4, 2, 1};" << endl;
      user << "  register unsigned long __key1=0, __key2=0, __sum=0, __s1=0, __s2=16;" << endl;
      for (int n=0; n<elements; n++) {
	if (n) {
	  user << "  __s1 += 3; __s1 %= 29;" << endl;
	  user << "  __s2 += 5; __s2 %= 29;" << endl;
	}
	if (safeHash) {
	  user << "  if (__s1 > 16 && " << instance(n) << " > ___p[__s1-16]) __s1 %= 17;" << endl;
	  user << "  if (__s2 > 16 && __sum > ___p[__s2-16]) __s2 %= 17;" << endl;
	}
        user << "  __key1 ^= (" << instance(n) << " << __s1);" << endl;
        user << "  __sum += " << instance(n) << ";" << endl;
        user << "  __key2 ^= (__sum << __s2);" << endl;
      }
    }
    user << "  return (__key1 ^ __key2);" << endl;
  }
  user << "}" << endl << endl;
}

void Parser::size(ostream &user)
{
  user << "long State::size() const {"<< endl << "    return (long) ( (unsigned char *) &__kids - (unsigned char *) this );" << endl << "}" << endl << endl;
}

void Parser::fileIO(ostream &user) 
{
  user << "int State::input(FILE *__fp) {" << endl << "  return fread(this, size(), 1, __fp);" << endl << "}" << endl << endl;
  user << "int State::output(FILE *__fp) const {" << endl << "  return fwrite(this, size(), 1, __fp);" << endl << "}" << endl << endl;
}

void Parser::access(ostream &user)
{
  user << "Access::Access() {" << endl << "  __current = new State;" << endl << "}" << endl << endl;
  user << "Access::~Access() {" << endl << "  delete __current;" << endl << "}" << endl << endl;
  user << "void Access::initial() {" << endl << "  __current->initial();" << endl << "}" << endl << endl;
  user << "int Access::isEnabled(int *__enabled) const {" << endl << "  return __current->isEnabled(__enabled);" << endl << "}" << endl << endl;
  user << "int Access::isTarget() const {" << endl << "  return __current->isTarget();" << endl << "}" << endl << endl;
  user << "int Access::isExcluded() const {" << endl << "  return __current->isExcluded();" << endl << "}" << endl << endl;
  //  user << "complex<double> Access::sojournTime(int __t, complex<double> __s) const {" << endl << "  return __current->sojournTime(__t, __s);" << endl << "}" << endl << endl;  user << "double Access::distribution(int __t) const {" << endl << "  return __current->distribution(__t);" << endl << "}" << endl << endl;
  user << "double Access::passageWeight() const {" << endl << "  return __current->passageWeight();" << endl << "}" << endl << endl;
  user << "void Access::fire(int __t, Access *next) const {" << endl << "  __current->fire(__t, next->__current);" << endl << "}" << endl << endl;
  user << "void Access::computeStateMeasures(double *__measures) const {" << endl << "  __current->computeStateMeasures(__measures);" << endl << "}" << endl << endl;
  user << "void Access::computeCountMeasures(int *__enabled, int __numberEnabled, double *__measures) const {" << endl << "  __current->computeCountMeasures(__enabled, __numberEnabled, __measures);" << endl << "}" << endl << endl;
  user << "double Access::getTransWeight(int __t) const {" << endl << "  return __current->getTransWeight(__t);" << endl << "}" << endl << endl;
// ### user  << Access::checkInvariants()
  user << "int Access::input(FILE *__fp) {" << endl << "  return __current->input(__fp);" << endl << "}" << endl << endl;
  user << "int Access::output(FILE *__fp) {" << endl << "  return __current->output(__fp);" << endl << "}" << endl << endl;
  user << "void Access::output() {" << endl << "  __current->output();" << endl << "}" << endl << endl;
  user << "long Access::size() {" << endl << "  return __current->size();" << endl << "}" << endl << endl;
  //  user << "complex<double> Access::exponential(double __lambda, complex<double> __s) const {\n  return __current->exponential(__lambda, __s);\n}" << endl << endl;
  //  user << "complex<double> Access::uniform(double __a, double __b, complex<double> __s) const {\n  return __current->uniform(__a, __b, __s);\n}" << endl << endl;
  //  user << "complex<double> Access::erlang(double __lambda, int __n, complex<double> __s) const {\n  return __current->erlang(__lambda, __n, __s);\n}" << endl << endl;
  //  user << "complex<double> Access::gamma(double __lambda, double __n, complex<double> __s) const {\n  return __current->gamma(__lambda, __n, __s);\n}" << endl << endl;
  //  user << "complex<double> Access::markov(complex<double> __s) const {\n  return __current->markov(__s);\n}" << endl << endl;
  //  user << "complex<double> Access::deterministic(double __d, complex<double> __s) const {\n  return __current->deterministic(__d, __s);\n}" << endl << endl;
  //  user << "complex<double> Access::immediate() const { return __current->immediate(); }" << endl << endl;

  user << "void Access::computeHashKeys() { return __current->computeHashKeys(); }" << endl << endl;
  user << "unsigned long Access::getPrimaryKey() { return __current->getPrimaryKey(); }" << endl << endl;
  user << "unsigned long Access::getSecondaryKey() { return __current->getSecondaryKey(); }" << endl << endl;
  user << "StateType Access::getType() { return __current->getType(); }" << endl;
  user << "int Access::findEnabledTransitions() { return __current->findEnabledTransitions(); }" << endl << endl;
  user << "int Access::getEnabledTransition(int __t) { return __current->getEnabledTransition(__t); }" << endl << endl;
  user << "void Access::fireEnabledTransition(int __t, Access *next) {" << endl << "  return __current->fireEnabledTransition(__t, next->__current);" << endl << "}" << endl << endl;
  user <<"int Access::numberChildren() { return __current->numberChildren(); }" << endl << endl;
  user << "int Access::getNumberEnabled() { return __current->getNumberEnabled(); }" << endl << endl;
  user << "void Access::setKids(int __k) { return __current->setKids(__k); }" << endl << endl;
  user << "double Access::getEnabledTransWeight(int __t) { return __current->getEnabledTransWeight(__t); }" << endl << endl;
  user << "StateMatrixElement& Access::getChild(int __t) { return __current->getChild(__t); }" << endl << endl;
  user << "double Access::rateOut() { return __current->rateOut(); }" << endl << endl;
  //  user << "double Access::average(int __n) const { return __current->average(__n); }" << endl << endl;
  //  user << "double Access::nth_moment(const complex<double> &__S, int __trans, int __n, complex<double> __delta_s) { return (__current->nth_moment(__S, __trans, __n, __delta_s)).real(); }" << endl << endl;
  user << "double Access::getTotalTransWeight() { return __current->getTotalTransWeight(); }" << endl << endl;
  user << "void Access::setChild(int __t, long __value, double __weight) { return __current->setChild(__t,__value,__weight); }" << endl << endl;
  user << "int Access::matrixOutput(FILE *__mf) { return __current->matrixOutput(__mf); }" << endl << endl;
  user << "void Access::calculateCountMeasures(double *__measures) { return __current->calculateCountMeasures(__measures); }" << endl << endl;
  user << "int Access::initialise(int __a, TransType *__b, int *__c) { return __current->initialise(__a,__b,__c); }" << endl << endl;
}

int Parser::generate(char *filename)
  {
  ofstream user(filename, ios::out);
  if (!user)
    terminate("hydra-generate.cxx cannot open file '%s' for writing", filename);
  headers(user);
  declarations(user);
  calcHasKeys(user);
  initialise(user);
  findEnabledTransitions(user);
  initial(user);
  isEnabled(user);
  isTarget(user);
  isExcluded(user);
  passageWeight(user);
  //  sojournTime(user);
  //  markov(user);
  //  distribution(user);
  calcHelpValues(user);
  checkInvariants(user);
  fire(user);
  getTransWeight(user);
  computeStateMeasures(user);
  computeCountMeasures(user);
  primaryHashKey(user);
  secondaryHashKey(user);
  size(user);
  output(user);
  fileIO(user);
  access(user);
  user.close();
  return 0;
}





