////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"

#define DIMENSION 5
#define STOP 4

#ifndef LOGGING
#define SILENT 1
#endif

/*
double SteadyStateAnalyser::calcAccuracy(Vector<double> &solution, Vector<double> &old) {
  double maxDiff = 0, solNorm = 0; 
  for (long n=0; n<tangible; n++) { 
    if (fabs(solution(n) - old(n)) > maxDiff) {
      maxDiff = fabs(solution(n) - old(n));
//      cout << "well, solution(" << n <<  ") = " << solution(n) << endl;
//      cout << "and old(" << n << ") = " << old(n) << endl;
//      cout << "so diff is (" << solution(n) -  old(n) << endl;
    }
    if (fabs(solution(n)) > solNorm) solNorm = fabs(solution(n));
  }
  cout << "maxDiff: " << maxDiff << " solNorm = " << solNorm << endl;
  cout << "accuracy: " << maxDiff/solNorm << endl;
  return maxDiff/solNorm;
}
*/
/*
void SteadyStateAnalyser::mycg(Vector <double> &solution, long maxIterations, double accuracy) {
  Vector<double> d, p;
  double alpha[DIMENSION];
  d.setLength(tangible);
  p.setLength(tangible);

  cout << "first" << endl;
  Vector<double> *r = new Vector<double>[DIMENSION];
  Vector<double> *x = new Vector<double>[DIMENSION];
  for (long n=0; n < DIMENSION; n++) {
    r[n].setLength(tangible);
    x[n].setLength(tangible);
  }

  for (n=0; n<tangible; n++)
    x[0][n] = solution(n);

  Q->calcResidual(x[0], r[0]);
  cout << "1st residual norm is " << r[0].infNorm() << endl;

  long iteration = 0;
  double convergence = 1.0;
  while (convergence > accuracy) {
    cout << "iteration " << iteration << endl;

    for (n=0; n<tangible; n++)
      d[n] = -r[0](n);
    Q->transMultiply(d, p);
    alpha[0] = (r[0].dot(p)/(r[0].dot(r[0])));
    for (n=0; n<tangible; n++) {
      x[1][n] = d(n) + alpha[0]*x[0](n);
//      cout << "x[" << n << "] = " << x[1](n) << endl;
    }
    Q->calcResidual(x[1], r[1]);
    x[1].normalise();
    convergence = calcAccuracy(x[1], x[0]);

    cout << "2nd residual norm is " << r[1].infNorm() << endl;
    //  cout << "DOT PRODUCT with 1 is = " << r[0].dot(r[1]) << endl;
    
    for (n=0; n<tangible; n++)
      d[n] = -r[1](n);
    Q->transMultiply(d, p);
    alpha[0] = (r[1].dot(p))/(r[1].dot(r[1]));
    alpha[1] = (r[0].dot(p))/(r[0].dot(r[0]));
    for (n=0; n<tangible; n++)
      x[2][n] = d(n) + alpha[0]*x[1](n) + alpha[1]*x[0](n);
    Q->calcResidual(x[2], r[2]);

    cout << "3rd residual norm is " << r[2].infNorm() << endl;
//    cout << "DOT PRODUCT with 1 is = " << r[0].dot(r[2]) << endl;
//    cout << "DOT PRODUCT with 2 is = " << r[1].dot(r[2]) << endl;

    for (n=0; n<tangible; n++)
    d[n] = -r[2](n);
    Q->transMultiply(d, p);
    alpha[0] = (r[2].dot(p))/(r[2].dot(r[2]));
    alpha[1] = (r[1].dot(p))/(r[1].dot(r[1]));
    alpha[2] = (r[0].dot(p))/(r[0].dot(r[0]));
    for (n=0; n<tangible; n++)
      x[3][n] = d(n) + alpha[0]*x[2](n) + alpha[1]*x[1](n) + alpha[2]*x[0](n);
    
    Q->calcResidual(x[3], r[3]);
    //  cout << "4th residual norm is " << r[3].infNorm() << endl;
    //  cout << "DOT PRODUCT with 1 is = " << r[0].dot(r[3]) << endl;
    //  cout << "DOT PRODUCT with 2 is = " << r[1].dot(r[3]) << endl;
    //  cout << "DOT PRODUCT with 3 is = " << r[2].dot(r[3]) << endl;
    
    for (n=0; n<tangible; n++)
      d[n] = -r[3](n);
    Q->transMultiply(d, p);
    alpha[0] = (r[3].dot(p))/(r[3].dot(r[3]));
    alpha[1] = (r[2].dot(p))/(r[2].dot(r[2]));
    alpha[2] = (r[1].dot(p))/(r[1].dot(r[1]));
    alpha[3] = (r[0].dot(p))/(r[0].dot(r[0]));
    for (n=0; n<tangible; n++)
      x[4][n] = d(n) + alpha[0]*x[3](n) + alpha[1]*x[2](n) + alpha[2]*x[1](n) + alpha[3]*x[0](n);
    x[3].normalise();
    x[4].normalise();
    convergence = calcAccuracy(x[4], x[3]);
    cout << "iteration = " << iteration << " accuracy = " << convergence << endl;
    
    Q->calcResidual(x[4], r[4]); 
    cout << "DOT PRODUCT with 1 is = " << r[0].dot(r[4]) << endl;
    cout << "DOT PRODUCT with 2 is = " << r[1].dot(r[4]) << endl;
    cout << "DOT PRODUCT with 3 is = " << r[2].dot(r[4]) << endl;
    cout << "DOT PRODUCT with 4 is = " << r[3].dot(r[4]) << endl;

    for (n=0; n<tangible; n++)
      x[0][n] = x[STOP](n);
    
    Q->calcResidual(x[0], r[0]);
    
    cout << "4th residual norm is " << r[0].infNorm() << endl;
    
    iteration += STOP;
  }

  for (n=0; n<tangible; n++)
    solution[n] = x[0](n);

  unScale(solution);
  solution.normalise();
  #ifndef RELEASE
  solution.textOutput("mycg.sol");
  #endif
}

*/

void SteadyStateAnalyser::tfqmr(Vector <double> &solution, long maxIterations, double accuracy) {
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting TFQMR solve..." << endl;
#endif

  Vector<double> w, y2n, y2n_1, v, d, rTilde, p1, p2;
  double tau, theta, eta, sigma, alpha, beta, rho1, rho2, c, 
    convergence = 1.0;
  register double factor, mulch, solNorm, diffNorm;
  int pass = 0;

  w.setLength(tangible);
  y2n.setLength(tangible);
  y2n_1.setLength(tangible);
  v.setLength(tangible);
  d.setLength(tangible);
  p1.setLength(tangible);
  p2.setLength(tangible);
  rTilde.setLength(tangible);

  Q->calcResidual(solution, w);
  long n, iterations;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
  for (n=0; n<tangible; n++)
    normSol[n] = solution[n];
  normSol.normalise();
  Q->calcResidual(normSol, normRes);
  cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
  cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif
  for (n=0; n<tangible; n++) {
    y2n_1[n] = w(n);
    rTilde(n) = w(n);
    d[n] = 0;
    p1[n] = v(n);
  }
  Q->transMultiply(y2n_1, v);
  for (n=0; n<tangible; n++)
    p1[n] = v(n);
// ### put me back! (think of twoNorm)
  tau = w.infNorm();
  theta = 0;
  eta = 0;
  rho1 = w.dot(rTilde);

  timer->read();
  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm bound   Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }

  for (iterations = 1; iterations < maxIterations; iterations++) {
    // (a)
    sigma = rTilde.dot(v);
    alpha = rho1/sigma;
    for (n=0; n<tangible; n++) {
      y2n(n) = y2n_1(n) - alpha*v(n);
      w[n] = w[n] - alpha*p1(n);
    }
   
    // (b) m = 2n - 1
//    Q->transMultiply(y2n_1, p1);
//    w[n] = w[n] - alpha*p1(n);
    mulch = theta*theta*eta/alpha;
    theta = w.infNorm()/tau;
    c = 1.0/sqrt(1.0+theta*theta);
    tau = tau*theta*c;
    eta = c*c*alpha;
    solNorm = 0; diffNorm = 0;
    for (n=0; n<tangible; n++) {
      d[n] = y2n_1(n) + mulch*d(n);
      factor = eta*d(n);
      if (fabs(factor) > diffNorm) diffNorm = fabs(factor);
      solution[n] = solution[n] + factor;
      //      if (fabs(solution(n) > solNorm)) solNorm = fabs(solution(n));

      //ASK WILL

      if (fabs(double(solution(n) > solNorm))) solNorm = fabs(solution(n));
    }
//    Q->calcResidual(solution, r);
#ifdef DEBUG
    cout << "iteration: " << iterations << " r.twoNorm < " << sqrt(iterations*2-1)*tau << endl;
#endif
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif
//    cout << "diffNorm = " << diffNorm << " solNorm = " << solNorm << endl;
//    convergence = diffNorm/solNorm;
//    cout << "accuracy = " << convergence << endl;
    if (convergence < accuracy) pass = 1;
    // ### check convergence - can do on basis of fabs(eta*d(n))
    // (b) m = 2n
    Q->transMultiply(y2n, p2);
    for (n=0; n<tangible; n++)
      w[n] = w[n] - alpha*p2(n);
    mulch = theta*theta*eta/alpha;
    theta = w.infNorm()/tau;
    c = 1/sqrt(1.0+theta*theta);
    tau = tau*theta*c;
    eta = c*c*alpha;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      d[n] = y2n(n) + mulch*d(n);
      factor = eta*d(n);
      if (fabs(factor) > diffNorm) diffNorm = fabs(factor);
      solution[n] = solution[n] + factor;
      if (fabs(solution(n)) > solNorm) solNorm = fabs(solution(n));
    }
//    Q->calcResidual(solution, r);
#ifdef DEBUG
    cout << "iteration: " << iterations << " r.twoNorm < " << sqrt(iterations*2)*tau << endl;
#endif
    convergence = diffNorm/solNorm;
#ifdef DEBUG
    cout << "accuracy = " << convergence << endl;
#endif
    if (convergence < accuracy && pass) break;
    pass = 0;

    // ### check convergence - can do on basis of fabs(eta*d(n))
    rho2 = rTilde.dot(w);
    beta = rho2/rho1;
    for (n=0; n<tangible; n++)
      y2n_1(n) = w(n) + beta*y2n(n);

    Q->transMultiply(y2n_1, p1);
    for (n=0; n<tangible; n++)
      v(n) = p1(n) + beta*(p2(n) + beta*v(n));

    if (solRepInterval && !(iterations % solRepInterval) && solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << sqrt(iterations*2.0)*tau << " | " << setw(17) << convergence << endl;
    }
    rho1 = rho2;
  }
  if (solRepInterval && solRepStyle != NONE) {
    cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << sqrt(iterations*2.0)*tau << " | " << setw(17) << convergence << endl;
  }
  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;

  if (convergence > accuracy)
    terminate("TFQMR failed to converge within %d iterations", maxIterations);

  cout.unsetf(ios::showpoint);

  solution.normalise();
  Q->calcResidual(solution, w);
  long status = checkVector(solution, w.infNorm());
  double timeUsed = timer->read();

  if (status & RANGE_VIOLATION) {
    refine(solution, maxIterations, accuracy);
    timeUsed += timer->read();
  } else 
    checkViolation(status);
  unScale(solution);
  solution.normalise();
  solution.output(steadyVectorFile);
  #ifndef RELEASE
  solution.textOutput("tfqmr.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): TFQMR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::update(Vector <double> &x, int k, Vector<double> *h, Vector<double> &s, Vector<double> *v) {
  Vector<double> y;

  long length = s.getLength();
  cout << "update1" << endl;
  y.setLength(length);
  long n;
  for (n=0; n<length; n++)
    y[n] = s(n);

  cout << "update2" << endl;
  int j;
  for (int i = k; i>=0; i--) {
    y[i] /= h[i](i);
    for (j = i-1; j>=0; j--)
      y[j] -= h[j](i) * y(i);
  }

  cout << "update3" << endl;
  double factor;
  for (j=0; j<=k; j++) {
    double factor = y(j);
    for (n=0; n<tangible; n++) {
      x(n) += v[j](n) * factor;
    }
  }
}
     
void SteadyStateAnalyser::gpr(double &dx, double &dy, double &cs, double &sn) {
  if (dy == 0.0) {
    cs = 1.0;
    sn = 0.0;
  } else if (fabs(dy) > fabs(dx)) {
    double temp = dx/dy;
    sn = 1.0 / sqrt(1.0 + temp*temp);
    cs = temp * sn;
  } else {
    double temp = dy/dx;
    cs = 1.0 / sqrt(1.0 + temp*temp);
    sn = temp*cs;
  }
}

void SteadyStateAnalyser::apr(double &dx, double &dy, double &cs, double &sn) {
  double temp = cs * dx + sn * dy;
  dy = -sn * dx + cs * dy;
//  cout << "apr dy = " << dy << endl;
  dx = temp;
//  cout << "apr dx = " << dx << endl;
}

void SteadyStateAnalyser::mycg(Vector <double> &solution, long maxIterations, double accuracy) {
  double resid, beta;
  int i,j = 1, k;
  long n;
  
  int m = 20;
  Vector<double> s, cs, sn, w, r, rr;
  s.setLength(m+1);
  cs.setLength(m+1);
  sn.setLength(m+1);

  w.setLength(tangible);
  r.setLength(tangible);
  rr.setLength(tangible);
  
  double normb = 1.0;
  Q->calcResidual(solution, r);
  beta = r.twoNorm();

  //  Vector<double> *v = new Vector<double>[m+1](tangible);
  Vector<double> *v = new Vector<double>[m+1];
  for(i=0; i<m+1; i++) v[i].setLength(tangible);
  //  Vector<double> *H = new Vector<double>[m+1](m);
  Vector<double> *H = new Vector<double>[m+1];
  for(i=0; i<m+1; i++) H[i].setLength(m);

  for (i=0; i<m+1; i++)
    for (n=0; n<m; n++)
      H[i][n] = 0;

//  cout << "1" << endl;

  while (j <= maxIterations) {
//    cout << "2" << endl;
    for (n=0; n<tangible; n++)
      v[0](n) = r(n)/beta;
    for (n=0; n<m+1; n++)
      s[n] = 0.0;
//    cout << "3" << endl;
    s[0] = beta;
    for (i=0; i<m && j <= maxIterations; i++, j++) {
//      cout << "4" << endl;
      Q->transMultiply(v[i], w);
//      cout << "w.infNorm = " << w.infNorm() << endl;
      for (k = 0; k <= i; k++) {
        H[k][i] = w.dot(v[k]);
        for (n=0; n<tangible; n++)
	  w(n) -= H[k][i] * v[k](n);
      }
//      cout << "5" << endl;
      H[i+1][i] = w.twoNorm();
//      cout << "hark! w.infNorm = " << w.infNorm() << endl;     
      for (n=0; n<tangible; n++)
	v[i+1](n) = w(n)/H[i+1][i];
//      cout << "v[" << i+1 << "].dot(v[0]) = " << v[i+1].dot(v[0]) << endl;
      for (k=0; k<i; k++)
	apr(H[k][i], H[k+1][i], cs[k], sn[k]);
      gpr(H[i][i], H[i+1][i], cs[i], sn[i]);
//      cout << "cs[" << i << "] = " << cs[i] << endl;
//      cout << "sn[" << i << "] = " << sn[i] << endl;
      apr(H[i][i], H[i+1][i], cs[i], sn[i]);
      apr(s[i], s[i+1], cs[i], sn[i]);
//      cout << "s[ " << i << " ] = " << s[i] << endl;
//      cout << "s[ " << i+1 << " ] = " << s[i+1] << endl;
      resid = fabs(s[i+1]);      
      cout << "iteration" << j << "resid1: " << resid << endl;
      if (resid <= 1e-10) {
	update(solution, i, H, s, v);
	solution.normalise();
	Q->calcResidual(solution, r);
	cout << "normalised resid = " << r.infNorm() << endl;
	unScale(solution);
	solution.normalise();
	solution.output(steadyVectorFile);
        #ifndef RELEASE
	solution.textOutput("gmres.sol");
	#endif
	delete [] v;
	return;
      }
    }
    cout << "continuing..." << endl;
    update(solution, m-1, H, s, v);
    
    Q->calcResidual(solution, r);
    beta = r.twoNorm();
    resid = beta;
    cout << "iteration" << j << "resid2: " << resid << endl;
    if (resid <= 1e-10) {
      update(solution, i, H, s, v);
      solution.normalise();
      Q->calcResidual(solution, r);
      cout << "normalised resid = " << r.infNorm() << endl;
      unScale(solution);
      solution.normalise();
      solution.output(steadyVectorFile);
      #ifndef RELEASE
      solution.textOutput("gmres.sol");
      #endif
      delete [] v;
      return;
    }
  }
  cout << "whoa collapse!!!" << endl;
}
