class Transition {
protected:
  char id[MAX_LENGTH];
  TransType type;
  int priority;
  CodeBlock condition, weight, action, sojourn, distribution;
public:
  char *getIdentity() { return id; }
  void setIdentity(char *source) { strcpy (id, source); }
  void setType(TransType _type) { type = _type; }
  void setPriority(int _priority) { priority = _priority; }
  TransType getType() const { return type; }
  int getPriority() { return priority; }
  CodeBlock &getCondition() { return condition; }
  CodeBlock &getWeight() { return weight; }
  CodeBlock &getAction() { return action; }
  CodeBlock &getSojourn() { return sojourn; }
  CodeBlock &getDistribution() { return distribution; }
};

class Constant {
protected:
  char id[MAX_LENGTH];
  CodeBlock define;
public:
  char *getIdentity() { return id; }
  void setIdentity(char *source) { strcpy (id, source); }
  CodeBlock &getDefine() { return define; }
};

class HelpValue {
protected:
  char id[MAX_LENGTH];
  char type[MAX_LENGTH];
  CodeBlock expression;
public:
  char *getIdentity() { return id; }
  void setIdentity(char *source) { strcpy (id, source); }
  char *getType() { return type; }
  void setType(char *source) { strcpy (type, source); }
  CodeBlock &getExpression() { return expression; }
};

class Invariant {
protected:
//  char id[MAX_LENGTH];
  CodeBlock expression;
public:
//  char *getIdentity() { return id; }
//  void setIdentity(char *source) { strcpy (id, source); }
  CodeBlock &getExpression() { return expression; }
};

class StateMeasure {
protected:
  char id[MAX_LENGTH];
  long estimator;
  CodeBlock expression;
public:
  StateMeasure() { estimator = 0; }
  void addEstimator(int est) { estimator ^= est; }
  int getEstimator() { return estimator; }
  char *getIdentity() { return id; }
  void setIdentity(char *source) { strcpy (id, source); }
  CodeBlock &getExpression() { return expression; }  
};

class CountMeasure {
protected:
  char id[MAX_LENGTH];
  long estimator;
  CodeBlock precondition, postcondition;
  int all, trans;
  FastDynaArray<int> transitionList;
public:
  CountMeasure() { estimator = 0; trans = 0; all = 0; }
  void addEstimator(int est) { estimator ^= est; }
  void addTransition(int t) {
    transitionList.check(trans);
    transitionList[trans] = t;
    trans++;
  }
  char *getIdentity() { return id; }
  int getEstimator() { return estimator; }
  void setIdentity(char *source) { strcpy (id, source); }
  int getAll() { return all; }
  void setAll(int _all) { all = _all; }
  CodeBlock &getPrecondition() { return precondition; }
  CodeBlock &getPostcondition() { return postcondition; }
  inline int getTrans() { return trans; }
  inline int getTransition(int t) { return transitionList[t]; }
};

class Parser {
protected:
  Tokeniser *tokeniser;
  Information *info;

  long maxStates, genRepInterval, solRepInterval, maxIterations;
  double maxCpuTime, accuracy, relaxParameter;
  int dynamic, advancedType, safeHash;
  double t_start, t_stop, t_step;
  int doSteady, doPassage, doTransient;

  ReportStyle genRepStyle, solRepStyle;
  SolutionMethod method;

  CodeBlock startVectorCB, initialCB, primaryCB, secondaryCB,
    includeCB, targetCB, outputCB, weightCB, excludedCB;

  char sourceFile[MAX_LENGTH], stateListFile[MAX_LENGTH], startVectorFile[MAX_LENGTH], steadyStateFile[MAX_LENGTH], transitionMatrixFile[MAX_LENGTH], resultsFile[MAX_LENGTH];

  int identifiers, constants, helpValues, transitions, invariants, stateMeasures,
countMeasures, elements;
  FastDynaArray<Transition *> trans;
  FastDynaArray<Constant *> cons;
  FastDynaArray<HelpValue *> helpv;
  FastDynaArray<Invariant *> invar;
  FastDynaArray<StateMeasure *> statem;
  FastDynaArray<CountMeasure *> countm;
  FastDynaArray<char *> store, type, instance;

  void addStateVectorElement(char *_type, char *_instance);
  int simpleType(char *_type);

  void destroy();
  void terminate(char *fmt, ...);

  int model();
  int passage();
  int stateVector(), instanceList(char *), transition(), constant(),
helpValue(), invariant();
  int generation();
  int solution(); 
  int performance();
  int stateMeasure(), countMeasure(), transitionList(CountMeasure *current);
  int output();
  int insert(char *);  
  int findTransitionIndex(char *target);
  void mark(char *);

  int generate(char *filename);
  void headers(ostream &user);
  void declarations(ostream &user);
  void calcHasKeys(ostream &user);
  void initialise(ostream &user);
  void findEnabledTransitions(ostream &user);
  void initial(ostream &user);
  void isEnabled(ostream &user);
  void isTarget(ostream &user);
  void isExcluded(ostream &user);
  void passageWeight(ostream &user);
  void fire(ostream &user);
  void calcHelpValues(ostream &user);
  void checkInvariants(ostream &user);
  void getTransWeight(ostream &user);
  void computeStateMeasures(ostream &user);
  void computeCountMeasures(ostream &user);	
  void output(ostream &user);
  void primaryHashKey(ostream &user);
  void secondaryHashKey(ostream &user);
  void size(ostream &user);
  void fileIO(ostream &user);
  void access(ostream &user);
  void sojournTime(ostream &user);
  void distribution(ostream &user);
  void markov(ostream &user);

  void setInformation();
  int nonTrivialCondition(CodeBlock &condition);
public:
  Parser(char*);
  ~Parser();
  int parse(char *filename, char* basename);
};




