#ifndef __VECTOR_HPP
#define __VECTOR_HPP

#include "dyna.hpp"

template <class T>
class Vector : public FastDynaArray<T>
{
  protected:
    int inputError(const char *filename);
    int outputError(const char *filename);
  public:
    Vector();
    Vector(long size);
    Vector(const Vector<T> &source);
    ~Vector();

    Vector<T> &operator=(const Vector<T> &source);
    Vector<T> operator-(const Vector<T> &source) const;
    Vector<T> operator+(const Vector<T> &source) const;

    int add(Vector &source) {
      for (long n=0; n<FastDynaArray<T>::length; n++)
        FastDynaArray<T>::array[n] += source(n);
      return 1;
    }

    int subtract(Vector &source) {
      for (long n=0; n<FastDynaArray<T>::length; n++)
        FastDynaArray<T>::array[n] -= source(n);
      return 1;
    }

    int addScaled(T scale, Vector &source) {
      for (long n=0; n<FastDynaArray<T>::length; n++) 
        FastDynaArray<T>::array[n] += scale*source(n);
      return 1;
    }

    int subtractScaled(T scale, Vector &source) {
      for (long n=0; n<FastDynaArray<T>::length; n++)
        FastDynaArray<T>::array[n] -= scale*source(n);
      return 1;
    }

    T dot(Vector &source) {
      T sum = 0;
      for (long n=0; n<FastDynaArray<T>::length; n++)
        sum += FastDynaArray<T>::array[n]*source(n);
      return (T) sum;
    }

    T oneNorm();
    T twoNorm();
    T infNorm();

    int multiply(T factor);
    int divide(T factor);
    int normalise();
    int normalise(long n);
/*    inline int normalise() {
      T factor = oneNorm();
      if (factor == 0) {
         #ifndef SILENT
         cout << "Vector normalise: Warning! Attempted division by zero\n";
         #endif
         return 0;
      }
      return multiply( (T) 1.0/factor);
    }*/
    
    int input(const char *filename);
    int output(const char *filename);

    int textInput(const char *filename);
    int textOutput(const char *filename, int precision=20);
};
#endif






